////////////////////////////////////////////////////////////////////////////////
//
// history:
//
// 2020-03-14  uj  1.0.0  setup
//
////////////////////////////////////////////////////////////////////////////////

/**
 * (c) Janicke Consulting, Germany 2022
 * mail info@janicke.de
 * 
 * Class to read CSV tables.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package de.janicke.airport.eedb;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.util.ArrayList;


public class CST {
  
    public static final String VERSION = "1.0.0";
    private enum TYPE { TABLE, PARAMETER };
    
    private int nrows;
    private int ncols;
    private String[] colname;
    private ArrayList<String[]> entry;
    private TYPE type;
    public String serror;
    
    /**
     * Read a csv table from the specified file with the specified section
     * and column separator
     * @param br       buffered reader
     * @param section  section (without []), or null
     * @param sep      file separator char (should be ; or ,)
     */
    public CST(BufferedReader br, String section, char sep) {
      try {
        serror = "";
        if (br == null)
          throw new Exception("buffered reader not reeady");
        read(br, section, sep);  
      }
      catch (Exception e) {
        e.printStackTrace(System.out);
        serror = e.getMessage();
        nrows = -1;
        ncols = -1;
      }
      finally {
        try { br.close(); } catch (Exception ee) {}
      }
    }

    /**
     * Read a csv table from the specified file with the specified section
     * and column separator
     * @param fin      file including full path information
     * @param section  section (without []), or null
     * @param sep      file separator char (should be ; or ,)
     */
    public CST(File fin, String section, char sep) {
      BufferedReader br = null;
      try {
        serror = "";
        if (fin == null || !fin.exists())
          throw new Exception("file does not exist:"+fin.getAbsolutePath());
        br = new BufferedReader(new FileReader(fin));
        read(br, section, sep);  
      }
      catch (Exception e) {
        e.printStackTrace(System.out);
        serror = e.getMessage();
        nrows = -1;
        ncols = -1;
      }
      finally {
        try { br.close(); } catch (Exception ee) {}
      }
    }
    

    private void read(BufferedReader br, String section, char sep) throws Exception {
      String line;
      boolean read = false;
      entry = new ArrayList<String[]>();
      nrows = -1;
      while ((line = br.readLine()) != null) {
        //
        char[] str = line.toCharArray();
        boolean inside = false;
        for (int i=0; i<str.length; i++) {
          if (str[i] == '"')
            inside = !inside;
          if (str[i] == sep && inside)
            str[i] = '#';
        }
        line = String.copyValueOf(str) + " ";
        //
        if (read && line.startsWith("["))
          break;
        if (!read) {
          if (section == null) {
            read = true;
            type = TYPE.TABLE;
          }
          else if (line.toLowerCase().startsWith("["+section.toLowerCase()+"]")) {
            read = true;
            type = (section.toLowerCase().startsWith("table")) ? TYPE.TABLE : TYPE.PARAMETER;
            continue;
          }
        }
        //
        // skip empty lines and comment lines
        if (!read || line.trim().length() == 0 || line.trim().startsWith("//"))
          continue;
        //
        String[] sa = line.split("["+sep+"]");
        StringBuilder sb = new StringBuilder();
        if (nrows == -1) {
          if (type == TYPE.TABLE) {
            ncols = sa.length;
            colname = new String[ncols];
            for (int i=0; i<sa.length; i++) {
              String s = sa[i].trim();
              sb.setLength(0);
              for (char c : s.toCharArray())
                if (c != ' ') sb.append(c);
              colname[i] = sb.toString();
            }
            nrows++;
            continue;   
          }
          else {
            nrows++; 
          }
        }
        if (type == TYPE.TABLE && sa.length != ncols)
          break;
        if (type == TYPE.PARAMETER && sa.length < 2)
          break;
        for (int i=0; i<sa.length; i++)
          sa[i] = sa[i].trim().replace("\"", "");
        entry.add(sa);
        nrows++;
      }
    }
    
    /**
     * Get row index (starting from 0) where the first entry (first column)
     * has the value "name"
     * @param name first entry 
     * @return row index or -1
     */
    public int getRow(String name) {
      int row = -1;
      for (int i=0; i<nrows; i++) {
        if (getString(i, 0).equalsIgnoreCase(name)) {
          row = i;
          break;
        }
      }
      return row;
    }
    
    /**
     * Get double at specified row  (PARAMETER)
     * @param row
     * @return 
     */
    public double getDouble(int row) {
      String s = getString(row, 1);
      double d = Double.NaN;
      try {
        d = Double.parseDouble(s.trim().replace(',', '.'));
      }
      catch (Exception e) {
        d = Double.NaN;
      }
      return d;
    }
    
    /**
     * Get int at specified row and column
     * @param row
     * @param col
     * @return 
     */
    public int getInt(int row, int col) {
      String s = getString(row, col);
      int i = Integer.MIN_VALUE;
      try {
        i = Integer.parseInt(s.trim());
      }
      catch (Exception e) {
        i = Integer.MIN_VALUE;
      }
      return i;
    }
    
    /**
     * Get double at specified row and column
     * @param row
     * @param col
     * @return 
     */
    public double getDouble(int row, int col) {
      String s = getString(row, col);
      double d = Double.NaN;
      try {
        d = Double.parseDouble(s.trim().replace(',', '.'));
      }
      catch (Exception e) {
        d = Double.NaN;
      }
      return d;
    }
    
    /**
     * Get double at specified row string (first column) and column header
     * @param srow
     * @param scol
     * @return 
     */
    public double getDouble(String srow, String scol) {
      int row = getRow(srow);
      if (row < 0)
        return Double.NaN;
      String s = getString(row, scol);  
      double d = Double.NaN;
      try {
        d = Double.parseDouble(s.trim().replace(',', '.'));
      }
      catch (Exception e) {
        d = Double.NaN;
      }
      return d;
    }
    
    /**
     * Get double at specified row and column header
     * @param row
     * @param scol
     * @return 
     */
    public double getDouble(int row, String scol) {
      String s = getString(row, scol);  
      if (s != null && s.equalsIgnoreCase("null"))
        s = null;
      double d = Double.NaN;
      try {
        d = Double.parseDouble(s.trim().replace(',', '.'));
      }
      catch (Exception e) {
        d = Double.NaN;
      }
      return d;
    }
    
     /**
     * Get string at specified row string  (PARAMETER)
     * @param srow
     * @return 
     */
    public String[] getStrings(String srow) {
      int row = getRow(srow);
      if (row < 0)
        return null;
      else {
        String[] sa = entry.get(row);
        String[] saa = new String[sa.length-1];
        for (int i=0; i<saa.length; i++)
          saa[i] = sa[i+1];
        return saa;
      }
    }
    
    /**
     * Get string at specified row string  (PARAMETER)
     * @param srow
     * @return 
     */
    public String getString(String srow) {
      int row = getRow(srow);
      if (row < 0)
        return null;
      else
        return getString(row, 1);  
    }
    
    /**
     * Get int at specified row string  (PARAMETER)
     * @param srow
     * @return 
     */
    public int getInt(String srow) {
      int row = getRow(srow);
      if (row < 0)
        return Integer.MIN_VALUE;
      else
        return getInt(row, 1);  
    }
    
    /**
     * Get double at specified row string  (PARAMETER)
     * @param srow
     * @return 
     */
    public double getDouble(String srow) {
      int row = getRow(srow);
      if (row < 0)
        return Double.NaN;
      else
        return getDouble(row, 1);  
    }
    
    /**
     * Get double at specified row string (PARAMETER)
     * @param srow
     * @return 
     */
    public double[] getDoubles(String srow) {
      String[] sa = getStrings(srow);
      if (sa == null)
        return null;
      double[] da = new double[sa.length];
      try {
        for (int i=0; i<da.length; i++)
          da[i] = Double.parseDouble(sa[i].trim().replace(',', '.'));
      }
      catch (Exception e) {
        da = null;
      }
      return da;
    }
    
    /**
     * Get string at specified row string (first column) and column header
     * @param srow
     * @param scol
     * @return 
     */
    public String getString(String srow, String scol) {
      int row = getRow(srow);
      if (row < 0)
        return null;
      else
        return getString(row, scol);  
    }
 
    /**
     * Get string at specified row and column header
     * @param row
     * @param scol
     * @return 
     */
    public String getString(int row, String scol) {
      int col = getColumnIndex(scol);
      return (col < 0) ? null : getString(row, col);
    }
    
    /**
     * Get string at specified row and column
     * @param row
     * @param col
     * @return 
     */
    public String getString(int row, int col) {
      if (row < 0 || row >= nrows || col < 0)
        return null;
      if (col >= entry.get(row).length)
        return null;
      return entry.get(row)[col];
    }
    
    /**
     * Get number of rows (not including header line)
     * @return 
     */
    public int getRowCount() { return nrows; }

    /**
     * Get number of columns
     * @return 
     */
    public int getColumnCount() { return ncols; }
    
    /**
     * Get name of column
     * @param col column index
     * @return 
     */
    public String getColumnName(int col) {
      if (type != TYPE.TABLE || col < 0 || col >= ncols) 
        return null;
      else
        return colname[col];
    }
    
    /**
     * Get index of column
     * @param scol column name
     * @return 
     */
    public int getColumnIndex(String scol) {
      if (type != TYPE.TABLE || scol == null)
        return -1;
      int col = -1;
      for (int i=0; i<ncols; i++)
        if (colname[i].equalsIgnoreCase(scol)) {
          col = i;
          break;
        }
      return col;
    }

    
}