////////////////////////////////////////////////////////////////////////////////
//
// history:
//
// 2020-03-14  uj  1.0.0  setup
// 2021-06-01  uj  1.1.1  (1+byp) corrected
// 2022-04-24  uj  1.1.2  handle prr=0 (FOI data)
//
////////////////////////////////////////////////////////////////////////////////

/**
 * (c) Janicke Consulting, Germany 2022
 * mail info@janicke.de
 * 
 * Class to calculate the non-volatile mass and number emission indices
 * using different methods.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package de.janicke.airport.eedb;

import java.util.Locale;


public class PM {
  
  public static final String VERSION = "1.1.2";
  public static boolean DEBUG = false;
  
  public static enum METHOD { FOA3N, FOA4, FOA4USR, FOA4GC, FOA4DF }; 

  public static final String GMD = "GMD";
  public static final String CEE = "CEE";
  private static final double TASTD = 273.15+10.;   // default temperature in K
  private static final double PASTD = 101325.0;     // default pressure in Pa
  private static final double RGAS = 8.314472;      // gas constant in J/(mol K)
  private static final double MDRY = 28.96546e-3;   // molar mass of dry air in kg/mol
  private static final double RDRY = RGAS/MDRY;     // gas constant of dry air in J/(kg*K)
  private static final double GAMMA = 1.4;          // ratio of specific heats
  private static final double CEFF = 0.9;           // compressor efficiency
  private static final double CPA = 1005.;          // heat capacity at constant pressure in J/(kg*K)
  private static final double CPE = 1250.;          // heat capacity for combustion in J/(kg*K)
  private static final double LCV = 43.2e+6;        // lower calorific fuel value in J/kg
  
  //
  // method specific contants
  
  private static final double[] FOA3N_GMD = { 40.e-9, 30.e-9, 20.e-9, 15.e-9 };
  private static final double   FOA3N_GSD = 1.7;
  private static final double   FOA3N_RHO = 1.e+6;
 
  private static final double[] FOA4_SF_AA =   { 1.0, 1.0, 0.8, 0.3 };
  private static final double[] FOA4_SF_CF34 = { 1.0, 0.4, 0.3, 0.3 };
  private static final double[] FOA4_SF_TL =   { 1.0, 1.0, 0.6, 0.3 };
  private static final double[] FOA4_SF_DAC0 = { 1.0, 0.9, 0.3, 0.3 };
  private static final double[] FOA4_SF_DAC1 = { 0.3, 0.3, 0.3, 1.0 };
  private static final double[] FOA4_AFR = { 45.0, 51.0, 83.0, 106.0 };
  private static final double[] FOA4_EIPM56 = { 4.6, 3.8, 4.5, 11.3 };
  private static final double[] FOA4_EIHC56 = { 0.04, 0.05, 0.08, 1.83 };
  
  private static final double FSCDEFAULT = 0.00068;
  private static final double S4TOS6DEFAULT = 0.024;
  private double Fsc = FSCDEFAULT;
  private double S4toS6 = S4TOS6DEFAULT;
  
  private static final double[] FOA4_GMD = { 40.e-9, 40.e-9, 20.e-9, 20.e-9 };
  private static final double   FOA4_GSD = 1.8;
  private static final double   FOA4_RHO = 1.e+6;
  
  private static final double   FOA4GC_FAC = 5.08;
  private static final double   FOA4GC_EXP = 0.185;
  private static final double   FOA4GC_GSD = 1.8;
  private static final double   FOA4GC_RHO = 1.e+6;
  
  private static final double   FAM20_RHOP = 1.770e+6;
  private static final double   FAM20_KTEM = 1.621e-5;
  private static final double   FAM20_DTEM = 0.39;
 // private static final double   FAM20_KTEM = 0.0125;
 // private static final double   FAM20_DTEM = 0.8;
  private static final double   FAM20_DFM = 2.76;
  
  public double[] FOA4USR_gmd = { 40.e-9, 40.e-9, 20.e-9, 20.e-9 };
  public double[] FOA4USR_gsd = {  FOA4_GSD, FOA4_GSD, FOA4_GSD, FOA4_GSD };

  private METHOD method;
  private boolean uncorrected;
  
 /**
   * PM calculation untilities of a given methodology
   * @param m the methodology (PM.FOA3N, PM.FOA4, ...)
   */
  public PM(METHOD m) {
    method = m;
    uncorrected = false;
  }
  
  /**
   * PM calculation untilities of a given methodology
   * @param m the methodology (PM.FOA3N, PM.FOA4, ...)
   * @param uc true if nbo line loss correction is made
   */
  public PM(METHOD m, boolean uc) {
    method = m;
    uncorrected = uc;
  }
  
  /**
   * Display implemented methodologies
   * @return string with descriptions
   */
  public String[] getImplementedMethods() {
    METHOD[] ma = METHOD.values();
    String[] sa = new String[ma.length];
    for (int i=0; i<sa.length; i++)
      sa[i] = ma[i].toString();
    return sa; 
  }
  
  /**
   * Get a short description of a iven methodology
   * @param m methodology
   * @return 
   */
  public String getInfo(METHOD m) {
    if (m == METHOD.FOA3N)
      return String.format("PM.FOA3N - Doc 9889 1st ed + LASPORT (GMD %1.1f/%1.1f/%1.1f/%1.1f nm, GSD %1.2f, RHO %1.2f 1/cm3)",
        FOA3N_GMD[0]*1.e9, FOA3N_GMD[1]*1.e9, FOA3N_GMD[2]*1.e9, FOA3N_GMD[3]*1.e9, 
        FOA3N_GSD, FOA3N_RHO*1.e-6);
    else if (m == METHOD.FOA4)
      return String.format("PM.FOA4 - Doc 9889 2nd ed (GMD %1.1f/%1.1f/%1.1f/%1.1f nm, GSD %1.2f, RHO %1.2f 1/cm3)",
        FOA4_GMD[0]*1.e9, FOA4_GMD[1]*1.e9, FOA4_GMD[2]*1.e9, FOA4_GMD[3]*1.e9, 
        FOA4_GSD, FOA4_RHO*1.e-6);
    else if (m == METHOD.FOA4USR)
      return String.format("PM.FOA4 - Doc 9889 2nd ed + user GMD/GSD (GMD %1.1f/%1.1f/%1.1f/%1.1f nm, GSD %1.2f/%1.2f/%1.2f/%1.2f)",
        FOA4USR_gmd[0]*1.e9, FOA4USR_gmd[1]*1.e9, FOA4USR_gmd[2]*1.e9, FOA4USR_gmd[3]*1.e9, 
        FOA4USR_gsd[0], FOA4USR_gsd[1], FOA4USR_gsd[2], FOA4USR_gsd[3]);
    else if (m == METHOD.FOA4GC)
      return String.format("PM.FOA4GC - Doc 9889 2nd ed + GMD(C) (FAC %1.2f, EXP %1.3f, GSD %1.2f, RHO %1.2f 1/cm3)",
        FOA4GC_FAC, FOA4GC_EXP, FOA4GC_GSD, FOA4GC_RHO*1.e-6);
    else if (m == METHOD.FOA4DF)
      return String.format("PM.FOA4DF - Doc 9889 2nd ed + GMD(C) + FAM20 ("
        + "FAC %1.2f, EXP %1.3f, GSD %1.2f, RHO0 %1.2f 1/cm3, RHOP %1.2f 1/cm3, DTEM %1.3f, KTEM %1.2e, DFM %1.2f)",
        FOA4GC_FAC, FOA4GC_EXP, FOA4GC_GSD, FOA4GC_RHO*1.e-6, FAM20_RHOP*1.e-6, FAM20_DTEM, FAM20_KTEM, FAM20_DFM);
    else return "?";
  }
  
  /**
   * Determines the particle diameter from given mass and number EI or concentration
   * using the default formula
   * @param em   mass EI in g/kg or mass concentration in g/m³
   * @param en   number EI in 1/kg or number concentration in 1/m³
   * @param sig  GSD
   * @return diameter in m
   */
  public static double GetDiameter(double em, double en, double sig) {
    double s = Math.log(sig);
    double rho = 1.e6;  // density in g/m³
    double d = Math.pow( (em/en)/((Math.PI/6.)*rho*Math.exp(4.5*s*s)), 1./3.);
    return d;
  }
  
  /**
   * Determines the number EI or concentration from given mass EI or concentration and diameter
   * using the default formula
   * @param em   mass EI in g/kg or mass concentration in g/m³
   * @param d    diameter in m
   * @param sig  GSD
   * @return  number EI in 1/kg or number concentration in 1/m³
   */
  public static double GetNumber(double em, double d, double sig) {
    double s = Math.log(sig);
    double rho = 1.e6;  // density in g/m³
    double en = em/((Math.PI/6.)*rho*d*d*d*Math.exp(4.5*s*s));
    return en;
  }
 
  /**
   * Calvert method to estimate a smoke number in case a value is missing or negative;
   * this will work if at least one of the 5 smoke numbers is known.
   * @param sn     smoke numbers
   * @param snmax  maximum smoke number
   * @param uid    engine uid
   * @param eng    engine type
   * @param cmb    combustor type
   * @param mnf    manufacturer
   * @return true if a replacement has taken place, false otherwise
   */
  public boolean reviewSN(double[] sn, double snmax, String uid, String eng, String cmb, String mnf) {
    boolean modified = false;
    //
    //  selected applicable factors
    double[] sf = null;
    if (uid != null && uid.toLowerCase().contains("aa") && mnf != null && 
      mnf.toLowerCase().contains("avia") && mnf.toLowerCase().contains("gatel")) {
      sf = FOA4_SF_AA;
    }
    else if (uid != null && uid.toLowerCase().contains("tl") & mnf != null && 
      mnf.toLowerCase().contains("textron")) {
      sf = FOA4_SF_TL;
    }
    else if (eng != null && eng.toLowerCase().contains("cf34")) {
      sf = FOA4_SF_CF34;
    }
    else if (cmb != null && cmb.toLowerCase().contains("dac")) {
      sf = FOA4_SF_DAC1;
    }
    else {
      sf = FOA4_SF_DAC0;
    }
    //
    // estimate maximum smoke number if required
    if (Double.isNaN(snmax) || snmax <= 0) {
      for (int im : MOD.I) {
        if (!Double.isNaN(sn[im]) && sn[im] > 0) {
          snmax = sn[im] / sf[im];
          break;
        }
      }
    }
    if (Double.isNaN(snmax) || snmax < 0)
      snmax = 10.;
    //
    // review smoke numbers and replace where required
    for (int im : MOD.I) 
      if (Double.isNaN(sn[im]) || sn[im] < 0) {
        sn[im] = sf[im] * snmax;
        modified = true;
      }
    return modified; 
  }
  
  /**
   * Geometric mean diameter
   * @param m
   * @param cm mass concentration at engine exit in g/m3
   * @param eng engine to access performance parameters of the engine
   * @return diameter in m or Double.NaN
   */
  public double getGMD(MOD.MODE m, double cm, ENG eng) {
    double d = Double.NaN;
    int im = m.ordinal();
    if (method == METHOD.FOA3N) 
      d = FOA3N_GMD[im];
    else if (method == METHOD.FOA4)
      d = FOA4_GMD[im];
    else if (method == METHOD.FOA4USR)
      d = FOA4USR_gmd[im];
    else if (method == METHOD.FOA4GC || method == METHOD.FOA4DF) {
      double prr = (Double.isNaN(eng.PRR) || eng.PRR <= 0) ? 1. : eng.PRR; //-2022-04-24
      double byp = (Double.isNaN(eng.BYP) || eng.BYP < 0) ? 0. : eng.BYP;
      if (!eng.isMTF())
        byp = 0;
      double afr = FOA4_AFR[im];
      double[][] tpa = getTP(TASTD, PASTD, MOD.ICAOTHR[im], MOD.MACH[im], prr, afr);      
      double rhoa = tpa[1][1]/(RDRY*tpa[0][1]);
      double rho4 = tpa[1][4]/(RDRY*tpa[0][4]);
      double cc = cm*(1 + byp)*(rho4/rhoa);
      d = FOA4GC_FAC*Math.pow(cc*1.e6, FOA4GC_EXP);
      d *= 1.e-9; 
    }
    /*
    else if (method == FAM20) {
      double prr = (Double.isNaN(eng.PRR) || eng.PRR < 0) ? 1. : eng.PRR;
      double afr = 1./(0.0121*MOD.ICAOTHR[mode] + 0.008);
      double[][] tpa = getTP(TASTD, PASTD, MOD.ICAOTHR[mode], MOD.MACH[mode], prr, afr);
      double t4o2 = tpa[0][4]/tpa[0][2];
      double dloss = -3.;
      if (eng.isDAC())
        d = (t4o2 < 2.85) ? 26.33*t4o2 - 35.98 : 20.;
      else
        d = 2.588*t4o2*t4o2 - 5.372*t4o2 + 16.72;
      if (!EEDB.USE_UNCORRECTED)
        d += dloss;
      d *= 1.e-9;
    }
    */
    return d;
  }
  
  /**
   *  Returns temperature (K) and pressure (P) at the engine planes 2, 3, 4
   * @param ta        ambient temperature in K
   * @param pa        ambient pressure in Pa
   * @param fthrust   fraction of maximum thrust [0,1]
   * @param mach      Mach number
   * @param prr       pressure ratio
   * @param afr       air-to-fuel ratio
   * @return          temperature [0][] and pressure[1][]
   */
  private double[][] getTP(double ta, double pa, double fthrust, double mach, double prr, double afr) { 
    double p1, p2, p3, p4;
    double t1, t2, t3, t4;
    //
    // ambient
    p1 = pa;
    t1 = ta; 
    //
    // compressor inlet
    p2 = p1 * Math.pow(1. + (GAMMA-1.)*0.5*mach*mach, GAMMA/(GAMMA-1.));
    t2 = t1 * (1. + (GAMMA-1.)*0.5*mach*mach);
    //
    // combustor inlet
    p3 = p2 * (1. + (prr-1.)*fthrust);
    t3 = t2 * Math.pow(p3/p2, (GAMMA-1.)/(GAMMA*CEFF));
    //
    // combustor exit
    p4 = p3;
    t4 = (afr*CPA*t3 + LCV)/(CPE*(1. + afr));
    //
    double[][] tpa = new double[2][5];
    tpa[0][0] = Double.NaN;
    tpa[0][1] = t1;
    tpa[0][2] = t2;
    tpa[0][3] = t3;
    tpa[0][4] = t4;
    tpa[1][0] = Double.NaN;
    tpa[1][1] = p1;
    tpa[1][2] = p2;
    tpa[1][3] = p3;
    tpa[1][4] = p4;
    return tpa;
  }
  
  /**
   * Geometric standard deviation
   * @param m
   * @return standard deviation or Double.NaN
   */
  public double getGSD(MOD.MODE m) {
    double d = Double.NaN;
    if (method == METHOD.FOA3N) 
      d = FOA3N_GSD;
    else if (method == METHOD.FOA4)
      d = FOA4_GSD;
    else if (method == METHOD.FOA4USR)
      d = FOA4USR_gsd[m.ordinal()];
    else if (method == METHOD.FOA4GC || method == METHOD.FOA4DF)
      d = FOA4GC_GSD;
    return d;
  }
  
  /**
   * Particle density
   * @param gmd   mean diameter in m
   * @return density in g/m3 or Double.NaN
   */
  public double getRho(double gmd) {
    double d = Double.NaN;
    if (method == METHOD.FOA3N) 
      d = FOA3N_RHO;
    else if (method == METHOD.FOA4 || method == METHOD.FOA4USR)
      d = FOA4_RHO;
    else if (method == METHOD.FOA4GC)
      d = FOA4GC_RHO;
    else if (method == METHOD.FOA4DF)
      d = FAM20_RHOP * 1.0 * 
          Math.pow(FAM20_KTEM*Math.pow(gmd, FAM20_DTEM-1.), 3.-FAM20_DFM);
    return d;
  }
  
  /**
   * Gets the mass concentration at engine exit
   * @param m
   * @param sn smoke number
   * @param eng
   * @return mass concentration in g/m3
   */
  public double getMassCon(MOD.MODE m, double sn, ENG eng) {
    double cm = Double.NaN;
    if (method == METHOD.FOA3N) {
      if (!Double.isNaN(sn) && sn >= 0) {
        if (sn <= 30.0)
          cm = 0.0694 * Math.pow(sn, 1.234);
        else 
          cm = 0.0297*sn*sn - 1.802*sn + 31.94;
        //
        // from mg to g
        cm *= 1.e-3;
      }
    }
    else if (method == METHOD.FOA4 ||
             method == METHOD.FOA4USR || 
             method == METHOD.FOA4GC || 
             method == METHOD.FOA4DF) {
      //
      // SCOPE11, cm in µg/m³
      cm = 648.4*Math.exp(0.0766*sn)/(1 + Math.exp(-1.098*(sn - 3.064)));
      //
      // correction factor, instrument -> engine exit
      double byp = (Double.isNaN(eng.BYP) || eng.BYP < 0) ? 0. : eng.BYP;
      if (!eng.isMTF())
        byp = 0;
      double cr = Math.log((3.219*cm*(1+byp) + 312.5)/(cm*(1+byp) + 42.6));
      if (!uncorrected)
        cm *= cr;
      //
      // from µg to g
      cm *= 1.e-6;
    }
    return cm;
  }
  
  /**
   * Determines the mass emission index based ion the pre-selected method
   * @param m
   * @param cm    mass concentration at engine exit in g/m3
   * @param eng   engine performance data
   * @return emission index in g/kg or Double.NaN;
   */
  public double getMassEI(MOD.MODE m, double cm, ENG eng) {
    double ei = Double.NaN;
    double byp = (Double.isNaN(eng.BYP) || eng.BYP < 0) ? 0. : eng.BYP;
    if (!eng.isMTF())
      byp = 0;
    if (method == METHOD.FOA3N) {
      double q = 0.776*FOA4_AFR[m.ordinal()]*(1+byp) + 0.877;
      ei = cm * q;
    }
    else if (method == METHOD.FOA4 || 
             method == METHOD.FOA4USR || 
             method == METHOD.FOA4GC || 
             method == METHOD.FOA4DF) {
      double q = 0.777*FOA4_AFR[m.ordinal()]*(1+byp) + 0.767;
      ei = cm * q;  
    }
    return ei;
  }
  
  /**
   * Determines the number EI or concentration from given mass
   * @param em   mass in g/UNIT
   * @param d    diameter in m
   * @param sig  GSD
   * @return     number in 1/UNIT
   */
  public double getNumber(double em, double d, double sig) {
    return getNumber(em, d, sig, 1.e6);
  }
  
  /**
   * Determines the number EI or concentration from given mass
   * @param em   mass in g/UNIT
   * @param d    diameter in m
   * @param sig  GSD
   * @param rho  density in g/m³
   * @return     number  in 1/UNIT
   */
  public double getNumber(double em, double d, double sig, double rho) {
    double s = Math.log(sig);
    double en = Double.NaN;
    if (method == METHOD.FOA3N || 
        method == METHOD.FOA4 ||
        method == METHOD.FOA4USR ||
        method == METHOD.FOA4GC) {
      en = em/((Math.PI/6.)*rho*d*d*d*Math.exp(4.5*s*s));
    }
    else if (method == METHOD.FOA4DF) {
      en = em/((Math.PI/6.)*rho*d*d*d*Math.exp(4.5*s*s));
      double phi = 3*FAM20_DTEM + (1.-FAM20_DTEM)*FAM20_DFM;
      en *= Math.exp((9.-phi*phi)*0.5*s*s);
    }
    return en;
  }
  
  /**
   * Sets the EIs (mass and number) for the provided engine for all LTO modes 
   * using the specified method; the values are stored in separate
   * fields of the ENG map; fields NVPM and NVPN are not changed.
   * @param eng
   * @return 
   */
  public String setEI(ENG eng) {
    if (eng.ISRC != EEDB.SOURCE.ICAO.hashCode() &&
        eng.ISRC != EEDB.SOURCE.FOI.hashCode() && 
        eng.ISRC != EEDB.SOURCE.OTHER.hashCode())
      return null;
    //
    // adjust SN if necessary
    double snmax = eng.get(EEDB.SMOKENUMBER, EEDB.SNMAX);
    double[] sn = new double[MOD.I.length];
    for (MOD.MODE m : MOD.MODE.values())
      sn[m.ordinal()] = eng.get(EEDB.SMOKENUMBER, m.toString());
    boolean mod = reviewSN(sn, snmax, eng.UID, eng.ENG, eng.CMB, eng.MNF);
    for (MOD.MODE m : MOD.MODE.values())
      eng.set(EEDB.SMOKENUMBER, m.toString(), sn[m.ordinal()]);
    //
    // set EIs
    String uc = (uncorrected) ? EEDB.X_UNCORRECTED : "";
    String snumber = EEDB.NVPN+uc+"_"+method;
    String smass = EEDB.NVPM+uc+"_"+method;
    String sgmd = GMD+uc+"_"+method;
    String scnc = CEE+uc+"_"+method;
    for (MOD.MODE m : MOD.MODE.values()) {
      double cm = getMassCon(m, eng.get(EEDB.SMOKENUMBER, m.toString()), eng);
      double gmd = getGMD(m, cm, eng);
      double gsd = getGSD(m);
      double rho = getRho(gmd);
      double eim = getMassEI(m, cm, eng);
      double ein = getNumber(eim, gmd, gsd, rho);
      eng.set(smass, m.toString(), eim);
      eng.set(snumber, m.toString(), ein);
      eng.set(sgmd,m.toString(), gmd*1.e9);
      eng.set(scnc, m.toString(), cm);
      setPM10(eng, m);
    }
    return method.toString();
  }
  
  /**
   * Inserts the PM mass contributions according to FOA4.
   * This requires that the nvPM mass has been calculated before and that
   * the HC emission index is available
   * @param eng  engine
   * @param im   LTO mode (0 for TO to 3 for ID)
   */
  private void setPM10(ENG eng, MOD.MODE m) {
    String testuid = "1AA002";
    boolean testex = DEBUG && eng.UID.equalsIgnoreCase(testuid);
    double ei=0., eisum=0.;
    int im = m.ordinal();
    //
    // non-volatile
    ei = eng.get(EEDB.NVPM, m.toString());
    if (Double.isNaN(ei)) {
      // no measured value from XLS, use FOA4 by default for PM10
      ei = eng.get(EEDB.NVPM+"_"+METHOD.FOA4, m.toString());
    }
    eng.set(EEDB.PM10_NV, m.toString(), ei);
    eisum += ei;
    //
    // volatile sulphate
    ei = 1000. * Fsc * S4toS6 * 96.0 / 32.;
    if (testex) System.out.printf("   FSC=%10.4e, S4to6=%10.4e           ==> EIVS=%1.5f g/kg\n", Fsc, S4toS6, ei);
    eng.set(EEDB.PM10_VS, m.toString(), ei);
    eisum += ei;
    //
    // volatile organics
    double hc = eng.get(EEDB.HC, m.toString());
    if (Double.isNaN(hc) || hc < 0)
      hc = 0.;
    ei = (FOA4_EIPM56[im]/FOA4_EIHC56[im]) * 0.001 * hc;
    if (testex) System.out.printf("   EIPM56=%7.4f, EIHC56=%7.6f, EIHC=%7.4f ==> EIVO=%1.5f g/kg\n", FOA4_EIPM56[im], FOA4_EIHC56[im], hc, ei);
    eng.set(EEDB.PM10_VH, m.toString(), ei);  
    eisum += ei;
    //
    // total
    if (testex) System.out.printf("                                         ==> EIPM=%1.5f g/kg\n", eisum);
    eng.set(EEDB.PM10, m.toString(), eisum);
  }

  private void test1() {
    // test SN review
    double snmax = -1;
    double[] sn1 = { 10, 5, 3, 1 };
    boolean mod = reviewSN(sn1, snmax, null, null, null, null);
    System.out.printf("mod=%b, expected: false\n", mod);
    double[] sn21 = { 20, -1, 3, 1 };
    double[] sn22 = { 20, -1, 3, 1 };
    mod = reviewSN(sn22, snmax, null, null, null, null);
    for (MOD.MODE m : MOD.MODE.values())
      System.out.printf("%d: %4.1f --> %4.1f\n", m.ordinal(), sn21[m.ordinal()], sn22[m.ordinal()]);
    snmax = 20.;
    double[] sn3 = { Double.NaN, -1, Double.NaN, -1 };
    mod = reviewSN(sn3, snmax, null, null, null, null);
    System.out.printf("snmax=%1.1f\n", snmax);
    for (MOD.MODE m : MOD.MODE.values())
      System.out.printf("%d: %4.1f\n", m.ordinal(), sn3[m.ordinal()]);
  }

  private static void test3() {
    PM pm = new PM(PM.METHOD.FOA4);
    String[] sta = { "", "amb", "cmpr in", "cmb in", "cmb out" };
    double thr=0.85;
    double man = 0.4;
    double prr = 35.;
    double afr = 41.;
    double[][] tpa = pm.getTP(283.15, 101325., thr, man, prr, afr);
    System.out.printf("thrust=%1.3f, Ma=%1.2f, PRR=%1.1f, AFR=%1.1f\n", thr, man, prr, afr);
    System.out.printf("%10s; %10s; %10s\n", "Station", "T (K)", "P (Pa)" );
    for (int i=1; i<5; i++)
      System.out.printf("%10s; %10.1f; %10.1f\n", sta[i], tpa[0][i], tpa[1][i]);  
  }
  
  public static void main(String[] args) {
    Locale.setDefault(Locale.ENGLISH);
    try {
      test3();
    }
    catch (Exception e) {
      e.printStackTrace(System.out);
    }
    System.exit(0);
  }
  
}
